;;; Copyright (c) 2003-2008 by the authors.
;;;
;;; See LICENCE and AUTHORS for details.

(in-package :hu.dwim.wui)

(defun http-request (uri &key
                     (protocol :http/1.1)
                     (method :get)
                     force-ssl
                     parameters
                     content
                     (content-type "application/x-www-form-urlencoded")
                     (content-length nil content-length-provided-p)
                     form-data
                     cookies
                     basic-authorization
                     (user-agent :wui)
                     (accept "*/*")
                     proxy
                     proxy-basic-authorization
                     headers
                     (redirect-limit 5)
                     (redirect-methods '(:get :head))
                     auto-referer
                     keep-alive
                     (close t)
                     (external-format-out *drakma-default-external-format*)
                     (external-format-in *drakma-default-external-format*)
                     force-binary
                     want-stream
                     stream)
  (unless (member protocol '(:http/1.0 :http/1.1) :test #'eq)
    (error "Don't know how to handle protocol ~S." protocol))
  (setf uri (if (typep uri 'uri)
                (clone-uri uri)
                (parse-uri uri)))
  (unless (member method +http-methods+ :test #'eq)
    (error "Don't know how to handle method ~S." method))
  (unless (member (scheme-of uri) '("http" "https") :test #'string=)
    (error "Don't know how to handle scheme ~S." (scheme-of uri)))
  (when (and close keep-alive)
    (error "CLOSE and KEEP-ALIVE must not be both true."))
  (when (and (eq content :continuation) content-length)
    (error "CONTENT-LENGTH must be NIL if CONTENT is :CONTINUATION."))
  (when (and form-data (not (eq method :post)))
    (error "FORM-DATA makes only sense with POST requests."))
  ;; convert PROXY argument to canonical form
  (when proxy
    (when (atom proxy)
      (setf proxy (list proxy 80))))
  (bind ((has-file-parameter?
          (bind ((result #f))
            (iter (for (parameter-name . parameter-value) :in parameters)
                  (when (typep parameter-value 'pathname)
                    (setf result #t)
                    (unless (eq method :post)
                      (error "Don't know how to handle file parameter ~S, as this is not a POST request." parameter-name))))
            result)))
    (when (eq method :post)
      ;; create content body for POST unless it was provided
      (unless content
        (if (or form-data
                has-file-parameter?)
            (progn
              (bind ((boundary (concatenate-string "----------" (random-simple-base-string 50))))
                (setf content (make-form-data-function parameters boundary))
                (setf content-type (concatenate-string "multipart/form-data; boundary=" boundary)))
              (unless (or has-file-parameter?
                          content-length-provided-p)
                (setf content-length (or content-length t))))
            (progn
              (setf content (alist-to-url-encoded-string parameters external-format-out))
              (setf content-type "application/x-www-form-urlencoded")))))
    (bind ((http-stream nil)
           (must-close? #f)
           (done? #f))
      (unwind-protect
          (progn
            (bind ((host (or (and proxy (first proxy))
                             (host-of uri)))
                   (port (cond (proxy (second proxy))
                               ((port-of uri))
                               (t (default-port uri)))))
              (setf http-stream (or stream (trivial-sockets:open-stream host port
                                                                 :element-type 'octet))))
            (when (and (or force-ssl
                           (eq (scheme-of uri) :https))
                       ;; don't attach SSL to existing streams
                       (not stream))
              (setf http-stream
                    #+:allegro
                    (socket:make-ssl-client-stream http-stream)
                    #-:allegro
                    (cl+ssl:make-ssl-client-stream http-stream)))
            (cond (stream
                   (setf (flexi-stream-element-type http-stream) 'character
                         (flexi-stream-external-format http-stream) +latin-1+))
                  (t
                   (setf http-stream
                         (make-flexi-stream (make-chunked-stream http-stream)
                                            :external-format +latin-1+))))
            (labels ((write-http-line (fmt &rest args)
                       (format http-stream "~?~C~C" fmt args #\Return #\Linefeed))
                     (write-header (name value-fmt &rest value-args)
                       (write-http-line "~A: ~?" name value-fmt value-args)))
              (when (and parameters (not (eq method :post)))
                (setf (query-of uri)
                      ;; append parameters to existing query of URI
                      (format nil "~@[~A~]~:*~:[~;&~]~A"
                              (query-of uri)
                              (alist-to-url-encoded-string parameters external-format-out))))
              (when (eq method :options*)
                ;; special pseudo-method
                (setf method :options
                      (path-of uri) "*"
                      (query-of uri) nil))
              (write-http-line "~A ~A ~A"
                               (string-upcase method)
                               (cond (proxy (render-uri uri nil))
                                     (t (format nil "~A~@[?~A~]"
                                                (or (path-of uri) "/")
                                                (query-of uri))))
                               (string-upcase protocol))
              (write-header "Host" "~A~@[:~A~]" (host-of uri) (non-default-port uri))
              (when user-agent
                (write-header "User-Agent" (user-agent-string user-agent)))
              (when basic-authorization
                (write-header "Authorization" "Basic ~A"
                              (base64:string-to-base64-string
                               (format nil "~A:~A"
                                       (first basic-authorization)
                                       (second basic-authorization)))))
              (when (and proxy proxy-basic-authorization)
                (write-header "Proxy-Authorization" "Basic ~A"
                              (base64:string-to-base64-string
                               (format nil "~A:~A"
                                       (first proxy-basic-authorization)
                                       (second proxy-basic-authorization)))))
              (when accept
                (write-header "Accept" "~A" accept))
              (when cookie-jar
                ;; write all cookies in one fell swoop, so even Sun's
                ;; web server has a change to get it
                (when-let (cookies (loop for cookie in (cookie-jar-cookies cookie-jar)
                                         when (send-cookie-p cookie uri force-ssl)
                                         collect (cookie-name cookie) and
                                         collect (cookie-value cookie)))
                  (write-header "Cookie" "~{~A=~A~^; ~}" cookies)))
              (when keep-alive
                (write-header "Connection" "Keep-Alive"))
              (when close
                (setf must-close? close)
                (write-header "Connection" "close"))
              (loop for (name . value) in additional-headers
                    do (write-header name "~A" value))
              (when content
                (when content-type
                  (write-header "Content-Type" "~A" content-type))
                (when (and content-length
                           (not (or (typep content 'ub8-vector)
                                    (eq content :continuation))))
                  ;; CONTENT-LENGTH forces us to compute request body in RAM
                  (setf content
                        (with-output-to-sequence (buffer-stream)
                          (write-http-request-content content buffer-stream))))
                (when (and (or (not content-length-provided-p)
                               (eq content-length t))
                           (typep content 'ub8-vector))
                  (setf content-length (length content)))
                (if content-length
                    (write-header "Content-Length" "~D" content-length)
                    (write-header "Transfer-Encoding" "chunked")))
              ;; end of request headers
              (when *header-stream*
                (terpri *header-stream*))
              (format http-stream "~C~C" #\Return #\Linefeed)
              (force-output http-stream)
              (when (and content (null content-length))
                (setf (chunked-stream-output-chunking-p
                       (flexi-stream-stream http-stream)) t))
              (labels ((finish-request (content &optional continuep)
                         (write-http-request-content content http-stream external-format-out)
                         (when continuep
                           (force-output http-stream)
                           (return-from finish-request))
                         (setf (chunked-stream-output-chunking-p
                                (flexi-stream-stream http-stream)) nil)
                         (finish-output http-stream)
                         (multiple-value-bind (server-protocol status-code)
                             ;; loop until status is NOT 100
                             (loop for (server-protocol status-code nil)
                                   = (read-status-line http-stream *header-stream*)
                                   when (= status-code 100)
                                   ;; ignore headers sent until non-100 status is seen
                                   do (read-http-headers http-stream *header-stream*)
                                   until (/= status-code 100)
                                   finally (return (values server-protocol status-code)))
                           (bind ((headers (read-http-headers http-stream *header-stream*))
                                 body external-format-body)
                             (bind ((connections (header-value :connection headers)))
                               (when connections
                                 (setf connections (split-tokens connections)))
                               (when (or (member "close" connections :test #'string-equal)
                                         (not (or (and (eq protocol :http/1.1)
                                                       (eq server-protocol :http/1.1))
                                                  (member "Keep-Alive" connections
                                                          :test #'string-equal))))
                                 (setf must-close? #t)))
                             (when cookie-jar
                               (update-cookies (get-cookies headers uri) cookie-jar))
                             (when (and redirect-limit
                                        (member status-code +http-redirect-codes+)
                                        (member method redirect-methods))
                               (unless (or (eq redirect-limit t)
                                           (and (integerp redirect-limit)
                                                (plusp redirect-limit)))
                                 (cerror "Continue anyway."
                                         "Status code was ~A, but ~:[REDIRECT-LIMIT is ~S~;redirection limit has been exceeded~]."
                                         status-code (integerp redirect-limit) redirect-limit))
                               (when auto-referer
                                 (setf additional-headers (set-referer uri additional-headers)))
                               (bind ((location (header-value :location headers))
                                      (new-uri (merge-uris (cond ((or (null location)
                                                                      (zerop (length location)))
                                                                  (warn "Empty `Location' header, assuming \"/\".")
                                                                  "/")
                                                                 (t location))
                                                           uri))
                                      ;; can we re-use the stream?
                                      (old-server-p (and (string= (host-of new-uri)
                                                                  (host-of uri))
                                                         (eql (port-of new-uri)
                                                              (port-of uri))
                                                         (eq (scheme-of new-uri)
                                                             (scheme-of uri)))))
                                 (unless old-server-p
                                   (setf must-close? #t
                                         want-stream nil))
                                 ;; try to re-use the stream, but only
                                 ;; if the user hasn't opted for a
                                 ;; connection which is always secure
                                 (bind ((re-use-stream (and old-server-p
                                                           (not must-close?)
                                                           (not force-ssl))))
                                   ;; close stream if we can't re-use it
                                   (unless re-use-stream
                                     (ignore-errors
                                       (close http-stream)))
                                   (setf done? #t)
                                   (return-from http-request
                                     (apply #'http-request new-uri
                                            :redirect-limit (if (integerp redirect-limit)
                                                                (1- redirect-limit)
                                                                redirect-limit)
                                            :stream (and re-use-stream http-stream)
                                            :additional-headers additional-headers
                                            args)))))
                             (bind ((transfer-encodings (header-value :transfer-encoding headers))
                                   (content-encodings (header-value :content-encoding headers)))
                               (when transfer-encodings
                                 (setf transfer-encodings (split-tokens transfer-encodings)))
                               (when content-encodings
                                 (setf content-encodings (split-tokens content-encodings)))
                               (when (member "chunked" transfer-encodings :test #'equalp)
                                 (setf (chunked-stream-input-chunking-p
                                        (flexi-stream-stream http-stream)) t))
                               (unless force-binary
                                 (handler-case
                                     (multiple-value-bind (type subtype parameters)
                                         (get-content-type headers)
                                       (declare (ignore subtype))
                                       (when (and (string-equal type "text")
                                                  (null (set-difference transfer-encodings
                                                                        '("chunked" "identity")
                                                                        :test #'equalp))
                                                  (null (set-difference content-encodings
                                                                        '("identity")
                                                                        :test #'equalp)))
                                         (bind ((charset (parameter-value "charset" parameters))
                                                (name (cond (charset (intern (string-upcase charset)
                                                                             :keyword))
                                                            (t external-format-in))))
                                           (setf external-format-body
                                                 (make-external-format name :eol-style :lf)))))
                                   (error (condition)
                                     (warn "Problems determining charset ~
\(falling back to binary):~%~A"
                                           condition)))))
                             (cond (external-format-body
                                    (setf (flexi-stream-external-format http-stream)
                                          external-format-body))
                                   (t (setf (flexi-stream-element-type http-stream) 'octet)))
                             (when force-binary
                               (setf (flexi-stream-element-type http-stream) 'octet))
                             (unless (or want-stream (eq method :head))
                               (bind (trailers)
                                 (multiple-value-setf (body trailers)
                                     (read-body http-stream headers must-close? external-format-body))
                                 (when trailers
                                   (warn "Adding trailers from chunked encoding to HTTP headers.")
                                   (setf headers (nconc headers trailers)))))
                             (setf done? #t)
                             (values (cond (want-stream http-stream)
                                           (t body))
                                     status-code
                                     headers
                                     uri
                                     http-stream
                                     must-close?)))))
                (when (eq content :continuation)
                  (return-from http-request #'finish-request))
                (finish-request content))))
        ;; the cleanup form of the UNWIND-PROTECT above
        (when (and http-stream
                   (or (not done?)
                       (and must-close?
                            (not want-stream)))
                   (not (eq content :continuation)))
          (ignore-errors
            (close http-stream)))))))

(def function write-http-request (request &optional (*standard-output* *standard-output*))
  (bind ((bytes
          (with-output-to-sequence (buffer) ; can't use *standard-output* in with-output-to-sequence due to some cryptic error
            (bind ((*standard-output* buffer))
              (write-string (string (http-method-of request)))
              (write-char #\Space)
              (write-uri (uri-of request))
              (write-char #\Space)
              (write-string (http-version-of request))
              (write-crlf)
              (write-http-request-headers (headers-of request))))))
    (write-sequence bytes *standard-output*))
  request)

(def function write-http-request-headers (headers-alist &optional (*standard-output* *standard-output*))
  (iter (for (header-name . header-value) :in headers-alist)
        (write-string header-name)
        (write-string ": ")
        (write-http-request-content header-value *standard-output*)
        (write-crlf)))

(def generic write-http-request-content (content http-stream)
  (:documentation "Sends CONTENT to the stream HTTP-STREAM as part of the request body depending on the type of CONTENT.")
  (:method (content http-stream)
    (when content
      (error "Don't know how to write content ~S in WRITE-HTTP-REQUEST-CONTENT." content)))
  (:method ((content string) http-stream)
    (write-string content http-stream))
  (:method ((content integer) http-stream)
    (princ content http-stream))
  (:method ((content vector) http-stream)
    (check-type content ub8-vector)
    (write-sequence content http-stream))
  (:method ((content stream) http-stream)
    (assert (input-stream-p content))
    (assert (open-stream-p content))
    (copy-stream content http-stream :element-type '(unsigned-byte 8)))
  (:method ((content pathname) http-stream)
    (with-input-from-file (input content :element-type '(unsigned-byte 8))
      (write-http-request-content input http-stream)))
  (:method ((content function) http-stream)
    (funcall content http-stream)))

;;; Partially based on code copied from Drakma. See relevant copyright notice below.

;;; Copyright (c) 2006-2007, Dr. Edmund Weitz.  All rights reserved.
;;;
;;; Redistribution and use in source and binary forms, with or without
;;; modification, are permitted provided that the following conditions
;;; are met:
;;;
;;;   * Redistributions of source code must retain the above copyright
;;;     notice, this list of conditions and the following disclaimer.
;;;
;;;   * Redistributions in binary form must reproduce the above
;;;     copyright notice, this list of conditions and the following
;;;     disclaimer in the documentation and/or other materials
;;;     provided with the distribution.
;;;
;;; THIS SOFTWARE IS PROVIDED BY THE AUTHOR 'AS IS' AND ANY EXPRESSED
;;; OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
;;; WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
;;; ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
;;; DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
;;; DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
;;; GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
;;; INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
;;; WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
;;; NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
;;; SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
